package java.time.zone;

import java.time.LocalDate;
import java.time.Duration;
import java.util.Collections;
import java.util.Arrays;
import java.time.Instant;
import java.io.DataInput;
import java.io.IOException;
import java.io.DataOutput;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.concurrent.ConcurrentHashMap;
import java.util.Objects;
import java.util.List;
import java.util.concurrent.ConcurrentMap;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.io.Serializable;

public final class ZoneRules implements Serializable
{
    private static final long serialVersionUID = 3044319355680032515L;
    private static final int LAST_CACHED_YEAR = 2100;
    private final long[] standardTransitions;
    private final ZoneOffset[] standardOffsets;
    private final long[] savingsInstantTransitions;
    private final LocalDateTime[] savingsLocalTransitions;
    private final ZoneOffset[] wallOffsets;
    private final ZoneOffsetTransitionRule[] lastRules;
    private final transient ConcurrentMap<Integer, ZoneOffsetTransition[]> lastRulesCache;
    private static final long[] EMPTY_LONG_ARRAY;
    private static final ZoneOffsetTransitionRule[] EMPTY_LASTRULES;
    private static final LocalDateTime[] EMPTY_LDT_ARRAY;
    
    public static ZoneRules of(ZoneOffset obj, ZoneOffset obj2, List<ZoneOffsetTransition> obj3, List<ZoneOffsetTransition> obj4, List<ZoneOffsetTransitionRule> obj5) {
        Objects.requireNonNull(obj, "baseStandardOffset");
        Objects.requireNonNull(obj2, "baseWallOffset");
        Objects.requireNonNull(obj3, "standardOffsetTransitionList");
        Objects.requireNonNull(obj4, "transitionList");
        Objects.requireNonNull(obj5, "lastRules");
        return new ZoneRules(obj, obj2, obj3, obj4, obj5);
    }
    
    public static ZoneRules of(ZoneOffset obj) {
        Objects.requireNonNull(obj, "offset");
        return new ZoneRules(obj);
    }
    
    ZoneRules(ZoneOffset zoneOffset, ZoneOffset zoneOffset2, List<ZoneOffsetTransition> list, List<ZoneOffsetTransition> list2, List<ZoneOffsetTransitionRule> list3) {
        this.lastRulesCache = new ConcurrentHashMap();
        this.standardTransitions = new long[list.size()];
        (this.standardOffsets = new ZoneOffset[list.size() + 1])[0] = zoneOffset;
        for (int i = 0; i < list.size(); ++i) {
            this.standardTransitions[i] = ((ZoneOffsetTransition)list.get(i)).toEpochSecond();
            this.standardOffsets[i + 1] = ((ZoneOffsetTransition)list.get(i)).getOffsetAfter();
        }
        ArrayList<LocalDateTime> list4 = new ArrayList();
        ArrayList<ZoneOffset> list5 = new ArrayList();
        list5.add(zoneOffset2);
        for (ZoneOffsetTransition zoneOffsetTransition : list2) {
            if (zoneOffsetTransition.isGap()) {
                list4.add(zoneOffsetTransition.getDateTimeBefore());
                list4.add(zoneOffsetTransition.getDateTimeAfter());
            }
            else {
                list4.add(zoneOffsetTransition.getDateTimeAfter());
                list4.add(zoneOffsetTransition.getDateTimeBefore());
            }
            list5.add(zoneOffsetTransition.getOffsetAfter());
        }
        this.savingsLocalTransitions = (LocalDateTime[])list4.toArray(new LocalDateTime[list4.size()]);
        this.wallOffsets = (ZoneOffset[])list5.toArray(new ZoneOffset[list5.size()]);
        this.savingsInstantTransitions = new long[list2.size()];
        for (int j = 0; j < list2.size(); ++j) {
            this.savingsInstantTransitions[j] = ((ZoneOffsetTransition)list2.get(j)).toEpochSecond();
        }
        if (list3.size() > 16) {
            throw new IllegalArgumentException("Too many transition rules");
        }
        this.lastRules = (ZoneOffsetTransitionRule[])list3.toArray(new ZoneOffsetTransitionRule[list3.size()]);
        return;
    }
    
    private ZoneRules(long[] standardTransitions, ZoneOffset[] standardOffsets, long[] savingsInstantTransitions, ZoneOffset[] wallOffsets, ZoneOffsetTransitionRule[] lastRules) {
        this.lastRulesCache = new ConcurrentHashMap();
        this.standardTransitions = standardTransitions;
        this.standardOffsets = standardOffsets;
        this.savingsInstantTransitions = savingsInstantTransitions;
        this.wallOffsets = wallOffsets;
        this.lastRules = lastRules;
        if (savingsInstantTransitions.length == 0) {
            this.savingsLocalTransitions = ZoneRules.EMPTY_LDT_ARRAY;
        }
        else {
            ArrayList<LocalDateTime> list = new ArrayList();
            for (int i = 0; i < savingsInstantTransitions.length; ++i) {
                ZoneOffsetTransition o = new ZoneOffsetTransition(savingsInstantTransitions[i], wallOffsets[i], wallOffsets[i + 1]);
                if (o.isGap()) {
                    list.add(o.getDateTimeBefore());
                    list.add(o.getDateTimeAfter());
                }
                else {
                    list.add(o.getDateTimeAfter());
                    list.add(o.getDateTimeBefore());
                }
            }
            this.savingsLocalTransitions = (LocalDateTime[])list.toArray(new LocalDateTime[list.size()]);
        }
    }
    
    private ZoneRules(ZoneOffset zoneOffset) {
        this.lastRulesCache = new ConcurrentHashMap();
        (this.standardOffsets = new ZoneOffset[1])[0] = zoneOffset;
        this.standardTransitions = ZoneRules.EMPTY_LONG_ARRAY;
        this.savingsInstantTransitions = ZoneRules.EMPTY_LONG_ARRAY;
        this.savingsLocalTransitions = ZoneRules.EMPTY_LDT_ARRAY;
        this.wallOffsets = this.standardOffsets;
        this.lastRules = ZoneRules.EMPTY_LASTRULES;
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws InvalidObjectException {
        throw new InvalidObjectException("Deserialization via serialization delegate");
    }
    
    private Object writeReplace() {
        return new Ser((byte)1, this);
    }
    
    void writeExternal(DataOutput dataOutput) throws IOException {
        dataOutput.writeInt(this.standardTransitions.length);
        long[] standardTransitions = this.standardTransitions;
        for (int length = standardTransitions.length, i = 0; i < length; ++i) {
            Ser.writeEpochSec(standardTransitions[i], dataOutput);
        }
        ZoneOffset[] standardOffsets = this.standardOffsets;
        for (int length2 = standardOffsets.length, j = 0; j < length2; ++j) {
            Ser.writeOffset(standardOffsets[j], dataOutput);
        }
        dataOutput.writeInt(this.savingsInstantTransitions.length);
        long[] savingsInstantTransitions = this.savingsInstantTransitions;
        for (int length3 = savingsInstantTransitions.length, k = 0; k < length3; ++k) {
            Ser.writeEpochSec(savingsInstantTransitions[k], dataOutput);
        }
        ZoneOffset[] wallOffsets = this.wallOffsets;
        for (int length4 = wallOffsets.length, l = 0; l < length4; ++l) {
            Ser.writeOffset(wallOffsets[l], dataOutput);
        }
        dataOutput.writeByte(this.lastRules.length);
        ZoneOffsetTransitionRule[] lastRules = this.lastRules;
        for (int length5 = lastRules.length, n = 0; n < length5; ++n) {
            lastRules[n].writeExternal(dataOutput);
        }
    }
    
    static ZoneRules readExternal(DataInput dataInput) throws IOException, ClassNotFoundException {
        int int1 = dataInput.readInt();
        long[] array = (int1 == 0) ? ZoneRules.EMPTY_LONG_ARRAY : new long[int1];
        for (int i = 0; i < int1; ++i) {
            array[i] = Ser.readEpochSec(dataInput);
        }
        ZoneOffset[] array2 = new ZoneOffset[int1 + 1];
        for (int j = 0; j < array2.length; ++j) {
            array2[j] = Ser.readOffset(dataInput);
        }
        int int2 = dataInput.readInt();
        long[] array3 = (int2 == 0) ? ZoneRules.EMPTY_LONG_ARRAY : new long[int2];
        for (int k = 0; k < int2; ++k) {
            array3[k] = Ser.readEpochSec(dataInput);
        }
        ZoneOffset[] array4 = new ZoneOffset[int2 + 1];
        for (int l = 0; l < array4.length; ++l) {
            array4[l] = Ser.readOffset(dataInput);
        }
        int byte1 = dataInput.readByte();
        ZoneOffsetTransitionRule[] array5 = (byte1 == 0) ? ZoneRules.EMPTY_LASTRULES : new ZoneOffsetTransitionRule[byte1];
        for (int n = 0; n < byte1; ++n) {
            array5[n] = ZoneOffsetTransitionRule.readExternal(dataInput);
        }
        return new ZoneRules(array, array2, array3, array4, array5);
    }
    
    public boolean isFixedOffset() {
        return this.savingsInstantTransitions.length == 0;
    }
    
    public ZoneOffset getOffset(Instant instant) {
        if (this.savingsInstantTransitions.length == 0) {
            return this.standardOffsets[0];
        }
        long epochSecond = instant.getEpochSecond();
        if (this.lastRules.length > 0 && epochSecond > this.savingsInstantTransitions[this.savingsInstantTransitions.length - 1]) {
            ZoneOffsetTransition[] transitionArray = this.findTransitionArray(this.findYear(epochSecond, this.wallOffsets[this.wallOffsets.length - 1]));
            ZoneOffsetTransition zoneOffsetTransition = null;
            int i = 0;
            while (i < transitionArray.length) {
                zoneOffsetTransition = transitionArray[i];
                if (epochSecond < zoneOffsetTransition.toEpochSecond()) {
                    return zoneOffsetTransition.getOffsetBefore();
                }
                ++i;
            }
            return zoneOffsetTransition.getOffsetAfter();
        }
        int binarySearch = Arrays.binarySearch(this.savingsInstantTransitions, epochSecond);
        if (binarySearch < 0) {
            binarySearch = -binarySearch - 2;
        }
        return this.wallOffsets[binarySearch + 1];
    }
    
    public ZoneOffset getOffset(LocalDateTime localDateTime) {
        Object offsetInfo = this.getOffsetInfo(localDateTime);
        if ((offsetInfo instanceof ZoneOffsetTransition)) {
            return ((ZoneOffsetTransition)offsetInfo).getOffsetBefore();
        }
        return (ZoneOffset)offsetInfo;
    }
    
    public List<ZoneOffset> getValidOffsets(LocalDateTime localDateTime) {
        Object offsetInfo = this.getOffsetInfo(localDateTime);
        if ((offsetInfo instanceof ZoneOffsetTransition)) {
            return ((ZoneOffsetTransition)offsetInfo).getValidOffsets();
        }
        return Collections.singletonList((ZoneOffset)offsetInfo);
    }
    
    public ZoneOffsetTransition getTransition(LocalDateTime localDateTime) {
        Object offsetInfo = this.getOffsetInfo(localDateTime);
        return (offsetInfo instanceof ZoneOffsetTransition) ? ((ZoneOffsetTransition)offsetInfo) : null;
    }
    
    private Object getOffsetInfo(LocalDateTime key) {
        if (this.savingsInstantTransitions.length == 0) {
            return this.standardOffsets[0];
        }
        if (this.lastRules.length > 0 && key.isAfter(this.savingsLocalTransitions[this.savingsLocalTransitions.length - 1])) {
            ZoneOffsetTransition[] transitionArray = this.findTransitionArray(key.getYear());
            Object offsetInfo = null;
            int length = transitionArray.length;
            int i = 0;
            while (i < length) {
                ZoneOffsetTransition zoneOffsetTransition = transitionArray[i];
                offsetInfo = this.findOffsetInfo(key, zoneOffsetTransition);
                if (offsetInfo instanceof ZoneOffsetTransition || offsetInfo.equals(zoneOffsetTransition.getOffsetBefore())) {
                    return offsetInfo;
                }
                ++i;
            }
            return offsetInfo;
        }
        int binarySearch = Arrays.binarySearch(this.savingsLocalTransitions, key);
        if (binarySearch == -1) {
            return this.wallOffsets[0];
        }
        if (binarySearch < 0) {
            binarySearch = -binarySearch - 2;
        }
        else if (binarySearch < this.savingsLocalTransitions.length - 1 && this.savingsLocalTransitions[binarySearch].equals(this.savingsLocalTransitions[binarySearch + 1])) {
            ++binarySearch;
        }
        if ((binarySearch & 0x1) == 0x0) {
            LocalDateTime localDateTime = this.savingsLocalTransitions[binarySearch];
            LocalDateTime localDateTime2 = this.savingsLocalTransitions[binarySearch + 1];
            ZoneOffset zoneOffset = this.wallOffsets[binarySearch / 2];
            ZoneOffset zoneOffset2 = this.wallOffsets[binarySearch / 2 + 1];
            if (zoneOffset2.getTotalSeconds() > zoneOffset.getTotalSeconds()) {
                return new ZoneOffsetTransition(localDateTime, zoneOffset, zoneOffset2);
            }
            return new ZoneOffsetTransition(localDateTime2, zoneOffset, zoneOffset2);
        }
        else {
            return this.wallOffsets[binarySearch / 2 + 1];
        }
    }
    
    private Object findOffsetInfo(LocalDateTime localDateTime, ZoneOffsetTransition zoneOffsetTransition) {
        LocalDateTime dateTimeBefore = zoneOffsetTransition.getDateTimeBefore();
        if (zoneOffsetTransition.isGap()) {
            if (localDateTime.isBefore(dateTimeBefore)) {
                return zoneOffsetTransition.getOffsetBefore();
            }
            if (localDateTime.isBefore(zoneOffsetTransition.getDateTimeAfter())) {
                return zoneOffsetTransition;
            }
            return zoneOffsetTransition.getOffsetAfter();
        }
        else {
            if (!localDateTime.isBefore(dateTimeBefore)) {
                return zoneOffsetTransition.getOffsetAfter();
            }
            if (localDateTime.isBefore(zoneOffsetTransition.getDateTimeAfter())) {
                return zoneOffsetTransition.getOffsetBefore();
            }
            return zoneOffsetTransition;
        }
    }
    
    private ZoneOffsetTransition[] findTransitionArray(int i) {
        Integer value = Integer.valueOf(i);
        ZoneOffsetTransition[] array = (ZoneOffsetTransition[])this.lastRulesCache.get(value);
        if (array != null) {
            return array;
        }
        ZoneOffsetTransitionRule[] lastRules = this.lastRules;
        ZoneOffsetTransition[] array2 = new ZoneOffsetTransition[lastRules.length];
        for (int j = 0; j < lastRules.length; ++j) {
            array2[j] = lastRules[j].createTransition(i);
        }
        if (i < 2100) {
            this.lastRulesCache.putIfAbsent(value, array2);
            return array2;
        }
        return array2;
    }
    
    public ZoneOffset getStandardOffset(Instant instant) {
        if (this.savingsInstantTransitions.length == 0) {
            return this.standardOffsets[0];
        }
        int binarySearch = Arrays.binarySearch(this.standardTransitions, instant.getEpochSecond());
        if (binarySearch < 0) {
            binarySearch = -binarySearch - 2;
        }
        return this.standardOffsets[binarySearch + 1];
    }
    
    public Duration getDaylightSavings(Instant instant) {
        if (this.savingsInstantTransitions.length == 0) {
            return Duration.ZERO;
        }
        return Duration.ofSeconds((long)(this.getOffset(instant).getTotalSeconds() - this.getStandardOffset(instant).getTotalSeconds()));
    }
    
    public boolean isDaylightSavings(Instant instant) {
        return !this.getStandardOffset(instant).equals(this.getOffset(instant));
    }
    
    public boolean isValidOffset(LocalDateTime localDateTime, ZoneOffset zoneOffset) {
        return this.getValidOffsets(localDateTime).contains(zoneOffset);
    }
    
    public ZoneOffsetTransition nextTransition(Instant instant) {
        if (this.savingsInstantTransitions.length == 0) {
            return null;
        }
        long epochSecond = instant.getEpochSecond();
        if (epochSecond >= this.savingsInstantTransitions[this.savingsInstantTransitions.length - 1]) {
            if (this.lastRules.length == 0) {
                return null;
            }
            int year = this.findYear(epochSecond, this.wallOffsets[this.wallOffsets.length - 1]);
            ZoneOffsetTransition[] transitionArray = this.findTransitionArray(year);
            int length = transitionArray.length;
            int i = 0;
            while (i < length) {
                ZoneOffsetTransition zoneOffsetTransition = transitionArray[i];
                if (epochSecond < zoneOffsetTransition.toEpochSecond()) {
                    return zoneOffsetTransition;
                }
                ++i;
            }
            if (year < 999999999) {
                return this.findTransitionArray(year + 1)[0];
            }
            return null;
        }
        else {
            int binarySearch = Arrays.binarySearch(this.savingsInstantTransitions, epochSecond);
            if (binarySearch < 0) {
                binarySearch = -binarySearch - 1;
            }
            else {
                ++binarySearch;
            }
            return new ZoneOffsetTransition(this.savingsInstantTransitions[binarySearch], this.wallOffsets[binarySearch], this.wallOffsets[binarySearch + 1]);
        }
    }
    
    public ZoneOffsetTransition previousTransition(Instant instant) {
        if (this.savingsInstantTransitions.length == 0) {
            return null;
        }
        long epochSecond = instant.getEpochSecond();
        if (instant.getNano() > 0 && epochSecond < Long.MAX_VALUE) {
            ++epochSecond;
        }
        long n = this.savingsInstantTransitions[this.savingsInstantTransitions.length - 1];
        if (this.lastRules.length > 0 && epochSecond > n) {
            ZoneOffset zoneOffset = this.wallOffsets[this.wallOffsets.length - 1];
            int year = this.findYear(epochSecond, zoneOffset);
            ZoneOffsetTransition[] transitionArray = this.findTransitionArray(year);
            int i = transitionArray.length - 1;
            while (i >= 0) {
                if (epochSecond > transitionArray[i].toEpochSecond()) {
                    return transitionArray[i];
                }
                --i;
            }
            if (--year > this.findYear(n, zoneOffset)) {
                ZoneOffsetTransition[] transitionArray2 = this.findTransitionArray(year);
                return transitionArray2[transitionArray2.length - 1];
            }
        }
        int binarySearch = Arrays.binarySearch(this.savingsInstantTransitions, epochSecond);
        if (binarySearch < 0) {
            binarySearch = -binarySearch - 1;
        }
        if (binarySearch <= 0) {
            return null;
        }
        return new ZoneOffsetTransition(this.savingsInstantTransitions[binarySearch - 1], this.wallOffsets[binarySearch - 1], this.wallOffsets[binarySearch]);
    }
    
    private int findYear(long n, ZoneOffset zoneOffset) {
        return LocalDate.ofEpochDay(Math.floorDiv(n + (long)zoneOffset.getTotalSeconds(), 86400L)).getYear();
    }
    
    public List<ZoneOffsetTransition> getTransitions() {
        ArrayList<ZoneOffsetTransition> list = new ArrayList();
        for (int i = 0; i < this.savingsInstantTransitions.length; ++i) {
            list.add(new ZoneOffsetTransition(this.savingsInstantTransitions[i], this.wallOffsets[i], this.wallOffsets[i + 1]));
        }
        return Collections.unmodifiableList(list);
    }
    
    public List<ZoneOffsetTransitionRule> getTransitionRules() {
        return Collections.unmodifiableList(Arrays.asList(this.lastRules));
    }
    
    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if ((o instanceof ZoneRules)) {
            ZoneRules zoneRules = (ZoneRules)o;
            return Arrays.equals(this.standardTransitions, zoneRules.standardTransitions) && Arrays.equals(this.standardOffsets, zoneRules.standardOffsets) && Arrays.equals(this.savingsInstantTransitions, zoneRules.savingsInstantTransitions) && Arrays.equals(this.wallOffsets, zoneRules.wallOffsets) && Arrays.equals(this.lastRules, zoneRules.lastRules);
        }
        return false;
    }
    
    @Override
    public int hashCode() {
        return Arrays.hashCode(this.standardTransitions) ^ Arrays.hashCode(this.standardOffsets) ^ Arrays.hashCode(this.savingsInstantTransitions) ^ Arrays.hashCode(this.wallOffsets) ^ Arrays.hashCode(this.lastRules);
    }
    
    @Override
    public String toString() {
        return "ZoneRules[currentStandardOffset=" + this.standardOffsets[this.standardOffsets.length - 1] + "]";
    }
    
    static {
        EMPTY_LONG_ARRAY = new long[0];
        EMPTY_LASTRULES = new ZoneOffsetTransitionRule[0];
        EMPTY_LDT_ARRAY = new LocalDateTime[0];
    }
}
